/***********************************************************************
 *                                                                     *
 * Projekt           :  M-Link zu FrSky Telemetrie-Konverter           *
 *                                                                     *
 * Projekt,Version   :  Version 0.6                                    *
 *                                                                     *
 *                                                                     *
 * Compiler-Version  : Arduino 1.6.7                                   *
 *                                                                     *
 * Prozessor         : Teensy 3.2                                      *
 *                                                                     *
 * Unterprogramme    :                                                 *
 *                                                                     *
 *                                                                     *
 * zus. Dokument.    : Strucktogramm                                   *
 *                                                                     *
 * Hinweise          :                                                 *
 *                                                                     *
 *                                                                     *
 *Original erstellt von T.Pfaff für ACT                                *
 *Abgeaendert,Datum  :  D. Kammerer, den 26.-01.-2016                  *
 *                                                                     *
 * geaendert         : ##1                                             *
 *                     Verschieben des blinken lassens der Build_In    *
 *                     LED nach der Abfrage, ob überhaupt M-Link       *
 *                     Sensordaten gekommen sind.                      *
 *                     Zeile von:                                      *
 *                     sendRSSI((uint8_t)(Ua0_aus),(uint8_t)(Ua1_aus), *
 *                              (uint8_t)(rssi_aus));                  *
 *                     in                                              *
 *                     sendRSSI((uint8_t)(Ua0_aus),(uint8_t)(Ua1_aus), *
 *                              (uint8_t)(rssi));                      *
 *                     Hier wird direkt die empfangene RRSI vom        *
 *                     Empfaenger ausgegeben und nicht eine irgendwie  *
 *                     berechnete und gemittelte. Meine Versuche haben *
 *                     ergeben, daß dieser berechnete Wert selbst bei  *
 *                     schlechten enpfangsbedingungen noch 97 angezeigt* 
 *                     hat, obwohl der vom Empfaenger schon auf 60 oder* 
 *                     kleiner war.                                    *
 *                                                                     *
 *                                                                     *
 *                                                                     *
 * Unterschiede zur                                                    *
 * version           :                                                 *
 *                                                                     *
 *                                                                     *
 * Kurzbeschreibung  :                                                 *

 * MLink Tx -> Arduino Rx
 * Basis Inverter -> Arduino IO7
 * Inverter see schematic
 * Special Thanks to:
 * ------------------
 * M.sc. T.Pfaff Version 13.09.2015
 * MLink-Library  T.Pfaff
 * FrSky-protocol with support form Mike and Kilrah from OpenRCforums (many thanks!)
 * 
 * Using MLink rssi value (ACT) often it goes under 10% for a verry short time. Thats normal but 
 * every time it triggers a rssi-warning. So here two things are done:
 * 1. rssi is filterd in a ring buffer and liniarized up to 2km maximum range (on 100mW Tx-power it goes to 0 at 2km)
 * 2. the raw rssi is additional send on ID 0x3F
 * Using rssi for warning is not quiet sensible, because the retreve power is normaly less than the Tx power, so
 * if rssi goes kritical telemetry gives up long time before.
 *
 * There are one main different between MLink and FrSky-hub-protocol: MLink can send up to 15 values of same
 * value, FrSky only one - exept temperatures and voltages. So if there are more than one or two (temperatutre)
 * values of the same type, the first one in stream are used. In case of more than one voltage, the highest
 * and lowest will be send. If there is a voltage on MLink-sensor "1" (as it is every time by ACT-MLink)  
 * this voltage is iterpreted as the reciver voltage and it will send on FrSky-sensor A1. It will by not
 * included in min-max-analytics.
 *
 * known unsolved errors:
 * If send 3,6V sometimes it appears a wrong value (25,6V and 29,2V). It seems to be a bug in MLink-processing 
 */

//#define NODEBUG   // Change this to #define DEBUG if needed 
#define DEBUG  // Change this to #define NODEBUG if needed 


//*******************************************************
//** Hier alle benoetigten Header-Files mit einbinden  **
//*******************************************************

//************************************************
//** Hier die benoetigten Konstanten definieren **
//************************************************
const byte Ein = 1;
const byte Aus = 0;

#define RX         9 // RxD-Pin der Seriellen Schnittstelle, wird zum 
                     // Synchronisieren benutzt

#define CYCLETIME             200


#define PROTOCOL_HEADER       0x7E
#define PROTOCOL_TAIL         0x7E
#define PROTOCOL_RSSI         0xFE
#define PROTOCOL_PRIM         0xFD
#define PROTOCOL_STUF         0x7D
#define PROTOCOL_STUFX0       0x5E
#define PROTOCOL_STUFX1       0x5D        

//Byte in frame has value 0x7E is changed into 2 bytes: 0x7D 0x5E 
//Byte in frame has value 0x7D is changed into 2 bytes: 0x7D 0x5D 
#define PROTOCOL_HUB_HEADER       0x5E
#define PROTOCOL_HUB_TAIL         0x5E
#define PROTOCOL_HUB_STUF         0x5D
#define PROTOCOL_HUB_STUFX0       0x3E
#define PROTOCOL_HUB_STUFX1       0x3D 


// Data Ids  (bp = before decimal point; af = after decimal point)
// Official data IDs
#define ID_TEMPRATURE1        0x02  
#define ID_RPM                0x03  
#define ID_FUEL_LEVEL         0x04  
#define ID_TEMPRATURE2        0x05
#define ID_VOLT               0x06  
#define ID_VOLTHD_BP          0x3A  
#define ID_VOLTHD_AP          0x3B  
#define ID_ALTITUDE_BP        0x10  
#define ID_ALTITUDE_AP        0x21  
#define ID_GPS_SPEED_BP       0x11  
#define ID_GPS_SPEED_AP       0x19   
#define ID_COURSE_BP          0x14  
#define ID_COURSE_AP          0x1C 
#define ID_VOLTAGE_NEW        0x39  
#define ID_CURRENT            0x28  
#define ID_V_VARIO            0x30
// Freie Sensornummern die abgezeigt werden ab 0x20 0x3C  - 0x3F 
#define ID_RSSI_RAW           0x2f 

// Freie Sensornummern für Zellspannungen: 0x2A, 0x2B, 0x2C, 0x2D, 0x2E, 0x2F  wird abgezeigt Umrechnung = 25,5
#define ID_Zelle_1            0x2A 
#define ID_Zelle_2            0x2B 
#define ID_Zelle_3            0x2C 
#define ID_Zelle_4            0x2D 
#define ID_Zelle_5            0x2E 
#define ID_Zelle_6            0x2F 

// **************************************************************************************************
// ** Vorwaertsdeklaration der im Programm verwendeten Funktionen (manche Kompiler benoetigen das) **
// **************************************************************************************************
static void sendRSSI(uint8_t ad0,uint8_t ad1,uint8_t rssi);
static void serializeFrsky(uint8_t id0, int16_t a0);
void sendTemp1(int16_t Tempf);
void sendTemp2(int16_t Tempf);
void sendCurr(uint16_t Stromf);
void sendVolt(int16_t Uf);
void sendVoltHD(uint16_t Uf);
void sendVoltZelle(int16_t Uf,byte zelle);
void sendAlt(int16_t Hochf);
void sendSpeed(int16_t v_horf);
void sendRpm(uint16_t rpmf);
void sendKap(uint16_t Kapf);
void sendVario(int16_t v_steigf);
void sendKurs(uint16_t Grad_Kursf);
void sendRSSI_RAW(uint16_t Rssif);

byte Read_TX();
void EVO_Taste(void);
void MSB_mitschreiben();

//********************************************************
//** Hier die benoetigten globalen Variablen definieren **
//********************************************************

byte MSB_Lese_Status = 0;    // wird verwendet, um nicht unnuetz in einer Warteschleife zu 
                             // warten, bis das erste Zeichen von der Schnittstelle 
                             // gekommen wird.
byte temp;
int Toggle = 0;
byte Daten_Da = false;       

byte Sensor_Nummer = 0;	// Zaehler, zaehlt immer von 0 bis 15
						            // zum abfragen der Sensorwerte aus dem Array Sensor_daten[]

// Hier werden die Werte der Sensoren gespeicher
// Speicherplatz 0 --> Sensorwert der auf der EVO-Anzeige auf 
// Platz 0 angezeigt worden waehre usw.
int  Sensor_daten[16]   = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

// Zu Jedem Sensorwert gehoert auch seine Einheit
byte Sensor_Einheit[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};   

// Hier wird gespeichert, ob das Alarmbit des Sensorwertes 
// gesetzt ist: 255 --> Alarmbit gesetzt
//                0 --> Alarmbit nicht gesetzt
// Wird hier nicht ausgewertet und beachtet
byte Sensor_Alarm[16]   = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};  

// Variablen deklaration von T. Pfaff
// **********************************
    uint32_t lastCycleTime = 0;
    uint8_t  cycleNum = 0;
    int16_t  Temp1=0;
    int16_t  Temp2=0;
    int16_t  U=0;
    int16_t  Uzelle=0;
    int16_t  Utest=0;
    uint16_t Ua0=100;
    uint16_t Ua1=100;    
    uint16_t Uamin=100;
    uint16_t Umax=100;
    int32_t  U_aus=500;
    uint32_t Ua0_aus=500;
    uint32_t Ua1_aus=500;
    uint16_t Strom=0;
    int16_t  v_steig=0;
    int16_t  v_hor=0;
    uint16_t rpm=0;
    uint16_t Grad_Kurs=0;
    int16_t  Hoch=0;
    uint16_t Kap=0; 
    uint16_t rssi=0; 
    uint32_t rssi_aus=100; 
    uint16_t U_Zelle_1 = 0;
    uint16_t U_Zelle_2 = 0;
    uint16_t U_Zelle_3 = 0;
    uint16_t U_Zelle_4 = 0;
    uint16_t U_Zelle_5 = 0;
    uint16_t U_Zelle_6 = 0;
    uint16_t U_Zelle[6]= {0,0,0,0,0,0};

    byte U_index = 0;
    char Temp1_adresse = -1;
    char Temp2_adresse = -1;

// Flags die zur Abarbeitung gebraucht werden
// ******************************************
    bool u_b         = false;    
    bool umax_b      = false;    
    bool uzelle_b    = false;
    bool ua0_b       = false;
    bool ua1_b       = false;
    bool Strom_b     = false;
    bool v_steig_b   = false;
    bool v_hor_b     = false;
    bool rpm_b       = false;
    bool Temp1_b     = false;
    bool Temp2_b     = false;
    bool Temp1adr_b  = false;
    bool Temp2adr_b  = false;
    bool Grad_Kurs_b = false;
    bool Hoch_b      = false;
    bool Kap_b       = false;
    bool rssi_b      = false;
    bool U_Zelle_1_b = false;
    bool U_Zelle_2_b = false;
    bool U_Zelle_3_b = false;
    bool U_Zelle_4_b = false;
    bool U_Zelle_5_b = false;
    bool U_Zelle_6_b = false;
    bool U_Zelle_b[6]= {false, false, false, false, false, false};
  
//--------------------------------------------------------
void setup() 
{
    /**************************************
    ** Globale Variablen initialisieren **
    **************************************/
    MSB_Lese_Status = 0;

   // Build in LED als I AM ALIVE einschalten 
   pinMode(LED_BUILTIN, OUTPUT);
   digitalWrite(LED_BUILTIN,1);

	// Nach dem Start 5Sekunden warten und nichts machen, damit das 
	// Sendemodul seinen normalen Betrieb aufnehmen kann.
	delay(5000);	
	  
	// ********************************************************************
	// ** Hier auf den naechste Frame warten und dann aufsynchronisieren **
	// ********************************************************************
	pinMode(RX, INPUT);
	// Warten, bis die RX-Leitung M-Link auf Low geht. Denn dann koennte ein Frame anfangen.
	do 
	  {
		  temp = digitalRead(RX);
	  } while (temp == 1);
	// Nun 3ms warten, dann sind die 23Bytes des COM-Ports 
	// 115kBaud auf jedenfall sicher Uebertragen. 
	// 23Bytes dauern ca. 2    ms
	delay(4);	// Nach dieser Zeit muessten nun alle Bytes uebertragen sein.

// Falls erforderlich Debugging Daten zum Seriellen Terminal schicken Freigeben
#ifdef DEBUG
  Serial.begin(9600);     // Ausgabeport zum debuggen
#endif
     
	// So, jetzt muessten wir synchron zur Ausgabe der Sensorwerte sein und 
  // koennen die serielle Schnittstelle mit 115kBaud initialisieren und einschalten
  
  // Bitte hier Kommentar davor setzen, wenn mit Transistorinverter zur Taranis gearbeitet wird
  Serial1.begin(9600,SERIAL_8N1_TXINV);  // Bit-Stream Ausgang FrSky Sensor Daten
                                         // --> 9600Baud, 8Datenbits, Keine Parität, RX Polarity invertiert
                                         // --> KEIN Transistorinverter mehr notwendig
                                         
   // Bitte hier Kommentar entfernen, wenn mit Transistorinverter zur Taranis gearbeitet wird
//  Serial1.begin(9600);  // Bit-Stream Ausgang FrSky Sensor Daten
                         // --> 9600Baud, 8Datenbits, Keine Parität, RX Polarity nicht invertiert
                         // --> hier ist EIN Transistorinverter notwendig damit der Konverter 
                         //     von der Taranis verstanden wird
  
  Serial2.begin(115200);                 // Bit-Stream Eingang M-Link Sensor Daten
  Ua1 = 0;
}  

// Okay, jetzt sind wir bereit und es kann los gehen
// *************************************************
void loop() 
{
    byte      einheit=0;
    uint32_t  RMP_Temp=0;
    uint32_t  temp1;
    float     wert=0;

    
    MSB_mitschreiben();  // MLink Frame vom COM-Port des HF-Modules lesen und auswerten und zur
                         // weiteren Verwendung abspeichern in den Arrays
                         // Sensor_daten[16] fuer die Sensorwerte und 
                         // Sensor_Einheit[16] fuer die entsprechenden Einheiten der Werte

    // **********************************************************
    // ** Ab hier die empfangenen M-Link Sensordaten auswerten **
    // **********************************************************
    if (Daten_Da == true) // Nur auswerten, wenn auch M-Link Daten angekommen sind
    {
        // ##1 Wenn Telemetriedaten vom Empfaenger kommen Build-in LED blinken lassen
        Toggle++;
        if (Toggle >= 20000) {  // Alle 20.000 Umlaufe die Build-in LED Toggeln
            digitalWrite(LED_BUILTIN, !digitalRead(LED_BUILTIN)); // Look I am alive
            Toggle = 0;     // von vorne wieder anfangen
        }

        if(Sensor_Einheit[Sensor_Nummer] > 0)	// Ist im Array Sensoreinheit an der Stelle Sensor_Nummer
        {                                     // ueberhaupt eine Einheit eingetragen?
         wert    = Sensor_daten[Sensor_Nummer];   // Sensorwert zur weiteren Verarberitung aus dem Array lesen
         einheit = Sensor_Einheit[Sensor_Nummer]; // Die dazugehoerende Einheit brauchen wir aber auch noch dazu  
         
         if (einheit == 0x05)  {   // Ist  der M-Link Wert die Drehzahl, dann muessen wir folgendesa Machen:
              RMP_Temp = 0x8000 - Sensor_daten[Sensor_Nummer];
#ifdef DEBUG
    Serial.println(Sensor_daten[Sensor_Nummer]);
    Serial.println(RMP_Temp);
#endif
 
              if ( RMP_Temp == 0x8000) RMP_Temp = 0;
              RMP_Temp = RMP_Temp * 100 / 6;     
              wert = RMP_Temp;     
                      
         }
        }  
  
        switch(einheit) // Je nach dem welcher Wert die Einheit hat (Ampere, Volt Meter usw.) wird hier der 
        {               // Wert dann entsprechend verarbeitet.
          
          case 0x01:    // Ist die Einheit  Volt?
                  // Wir schauen nun nach, auf welchem Sensorplatz die Spannung gefunden wurde
                  // Auf dem M-Link Sensorplatz 0 muß immer die Empfängerspannung sei, die dann 
                  // als FrSky Sensor A1 ausgegen wird.
                  // Um auch den FrSky-Sensor A2 ansprechen zu koennen muss auf Sensor Platz 2 
                  // wieder eine Spannung kommen.
                  // Um auch den FrSky-Sensor VFAS ansprechen zu koennen muss auf Sensor Platz 3 
                  // auch wieder eine Spannung kommen.

                  switch (Sensor_Nummer)
                  {
                      case 0x00:                                // Hier jetzt Sensorplatz 0 bzw. A1 bearbeiten
                                Ua0    = (uint16_t)(wert*1.96);  
                                ua0_b  = true;                   // Flag A1 Spannungswert wurde berechnet.
                       break;
                       
                       case 0x02:                               // Hier jetzt Sensorplatz 2 bzw. A2 bearbeiten
                                Ua1    = (uint16_t)(wert*1.96); 
                                ua1_b  = true;                   // Flag A1 Spannungswert wurde berechnet.
                       break;

                       case 0x03:                               // Hier jetzt Sensorplatz 3 bzw. VFAS bearbeiten
                                wert = abs(wert);
                                Umax   = (int16_t)(wert/1.9);  
                                Umax   = abs(Umax);
                                umax_b = true;                 // Flag VFAS Spannungswert wurde berechnet.
                       break;

                       case 0x09:                               // Hier jetzt Sensorplatz 10 bzw. Zelle1 bearbeiten
                                wert = abs(wert);
                                U_Zelle_1   = (int16_t)(wert);  
                                U_Zelle_1_b = true;            // Flag Zelle1 Spannungswert wurde berechnet.
                       break;

                       case 0x0A:                               // Hier jetzt Sensorplatz 11 bzw. Zelle2 bearbeiten
                                wert = abs(wert);
                                U_Zelle_2   = (int16_t)(wert);  
                                U_Zelle_2_b = true;             // Flag Zelle2 Spannungswert wurde berechnet.
                       break;
                       
                       case 0x0B:                               // Hier jetzt Sensorplatz 12 bzw. Zelle3 bearbeiten
                                wert = abs(wert);
                                U_Zelle_3   = (int16_t)(wert);  
                                U_Zelle_3_b = true;             // Flag Zelle3 Spannungswert wurde berechnet.
                       break;
                       
                       case 0x0C:                               // Hier jetzt Sensorplatz 13 bzw. Zelle4 bearbeiten
                                wert = abs(wert);
                                U_Zelle_4   = (int16_t)(wert);  
                                U_Zelle_4_b = true;             // Flag Zelle4 Spannungswert wurde berechnet.
                       break;
                       
                       case 0x0D:                               // Hier jetzt Sensorplatz 14 bzw. Zelle5 bearbeiten
                                wert = abs(wert);
                                U_Zelle_5   = (int16_t)(wert);  
                                U_Zelle_5_b = true;             // Flag Zelle5 Spannungswert wurde berechnet.
                       break;
                       
                       case 0x0E:                               // Hier jetzt Sensorplatz 15 bzw. Zelle6 bearbeiten
                                wert = abs(wert);
                                U_Zelle_6   = (int16_t)(wert);  
                                U_Zelle_6_b = true;             // Flag Zelle6 Spannungswert wurde berechnet.
                       break;
                  }
          break; 
          case 0x02: Strom     = (uint16_t)(      wert); Strom_b     = true; break;
          case 0x03: v_steig   = ( int16_t)( 10.0*wert); v_steig_b   = true; break;
          case 0x05: rpm       = (uint16_t)( wert / 100); rpm_b       = true; break;
          case 0x06: 
                     Temp1      = ( int16_t)(      wert);  // Es wird im Moment nur ein Temperaturwert behandelt
                     Temp1_b    = true;
                break; 
          case 0x07: Grad_Kurs = (uint16_t)(100.0*wert); Grad_Kurs_b = true; break;
          case 0x08: Hoch      = ( int16_t)(100.0*wert); Hoch_b      = true; break;
          case 0x09: Kap       = (uint16_t)(      wert); Kap_b       = true; break;
          case 0x0A: rssi      = (uint8_t) (      wert); rssi_b      = true; break;
        }
    
		Sensor_Nummer++;		// Den naechsten Sensor begutachten
		if (Sensor_Nummer >= 16) Sensor_Nummer = 0;	// Sind wir mit allen 16 Sensorplaetzen durch,
													                      // dann fangen wir wieder von Vorne an.
		
    // *********************************************
    // ** Ab hier wird der FrSky-Stream generiert **
    // *********************************************
    temp1 = millis() - lastCycleTime;
    if (temp1 >= CYCLETIME) // Nur alle 200ms einen FrSky-Stream senden!!
    {
        lastCycleTime = millis(); // Neue Zykluszeit abspeichern
        cycleNum++;

        if(ua0_b || ua1_b || rssi_b)  // Ist ein A1, A2 oder ein RSSI-Wert gekommen?
        {                             // Ja, dann.....
            Ua0_aus  = Ua0;
            Ua1_aus  = Ua1;       
            
            // Output for remote voltage A1, A2 and link quality
//            sendRSSI((uint8_t)(Ua0_aus),(uint8_t)(Ua1_aus),(uint8_t)(rssi_aus));
            sendRSSI((uint8_t)(Ua0_aus),(uint8_t)(Ua1_aus),(uint8_t)(rssi));
//            sendRSSI_RAW(Ua0_xx);
            
            // A1, A2 und RSSI als gesendet markieren
            ua0_b  = false;
            ua1_b  = false;
            rssi_b = false;
        }
        
        if(umax_b) // Hier, wenn vorhanden, VFAS als User Daten senden
        {
            sendVoltHD((uint16_t)abs(Umax));
            umax_b = false;   // VFAS als gesendet markieren
        }

        if (U_Zelle_1_b)
        {
            sendZelle(ID_Zelle_1,(uint16_t)U_Zelle_1);
            U_Zelle_1_b = false;            
        }

        if (U_Zelle_2_b)
        {
            sendZelle(ID_Zelle_2,(uint16_t)U_Zelle_2);
            U_Zelle_2_b = false;            
        }

        if (U_Zelle_3_b)
        {
            sendZelle(ID_Zelle_3,(uint16_t)U_Zelle_3);
            U_Zelle_3_b = false;            
        }

        if (U_Zelle_4_b)
        {
            sendZelle(ID_Zelle_4,(uint16_t)U_Zelle_4);
            U_Zelle_4_b = false;            
        }

        if (U_Zelle_5_b)
        {
            sendZelle(ID_Zelle_5,(uint16_t)U_Zelle_5);
            U_Zelle_5_b = false;            
        }

        if (U_Zelle_6_b)
        {
            sendZelle(ID_Zelle_6,(uint16_t)U_Zelle_6);
            U_Zelle_6_b = false;            
        }

        if(Strom_b) // Hier, wenn vorhanden, Strom als User Daten senden
        {
          sendCurr(Strom);
          Strom_b = false;
        }
        if(v_steig_b) // Hier, wenn vorhanden, Vertikales Steigen als User Daten senden
        {
          sendVario(v_steig);
          v_steig_b = false;
        }
        if(v_hor_b)   // Hier, wenn vorhanden, horizontale Geschwindigkeit als User Daten senden
        {
          sendSpeed(v_hor);
          v_hor_b = false;
        }
        if(rpm_b)   // Hier, wenn vorhanden, Drehzahl als User Daten senden
        {
          sendRpm(rpm);
          rpm_b = false;
        }
        if(Temp1_b)   // Hier, wenn vorhanden, die Temperatur als User Daten senden
        {
          sendTemp1(Temp1);
          Temp1_b = false;
        }
        if(Grad_Kurs_b)   // Hier, wenn vorhanden, Kurs in Grad als User Daten senden
        {
          sendKurs(Grad_Kurs);
          Grad_Kurs_b = false;
        }
        if(Hoch_b)      // Hier, wenn vorhanden, Hoehe als User Daten senden
        {
          sendAlt(Hoch);
          Hoch_b = false;
        }
        // *****************************************************************
        // ** Hier, wenn vorhanden, Kappazitaet als User Daten senden     **
        // ** Wird zwar nicht gebraucht, da die Taranis aus dem Strom die **
        // ** Verbrauchte Kappazitaet selbst errechnen kann               **
        // *****************************************************************
        if(Kap_b)      
        {            
          sendKap(Kap);
          Kap_b = false;
        }
      }
    } 
}        
